# -*- coding: utf-8 -*-

"""Top-level package for gql_py."""

__author__ = """Paul Hallett"""
__email__ = 'paulandrewhallett@gmail.com'
__version__ = '0.5.0'

from .gql import Gql, to_camelcase, normalise_query  # noqa



# -*- coding: utf-8 -*-

"""Main module."""

import re
from collections import namedtuple
from typing import Any, Mapping, Optional

import requests

EXTRA_WHITE_SPACE = re.compile(r'\s{2,}')


def normalise_query(query: str) -> str:
    """
    Strip redundant white space from the query string to save space.
    """
    return EXTRA_WHITE_SPACE.sub(' ', query.strip())


def to_camelcase(variables: Mapping[str, Any]) -> Mapping[str, Any]:
    """
    Transform typical python underscore-style variables
    into camelCased variables that graphQL expects.
    NOTE: Converts nested dictionaries recursively so all keys are camelCased.
    """
    def convert(string):
        # thanks to https://stackoverflow.com/a/47253475/1327062
        return re.sub('_([a-zA-Z0-9])', lambda m: m.group(1).upper(), string)

    converted_dict = {}
    for key, value in variables.items():
        if isinstance(value, dict):
            value = to_camelcase(value)
        key = convert(key)
        converted_dict[key] = value
    return converted_dict


def prepared_payload(
    *,
    cleaned_query: str,
    camelcased_variables: Optional[Mapping[str, Any]],
) -> Mapping[str, Any]:
    return {
        'query': cleaned_query,
        'variables': camelcased_variables,
    }


GraphQLResponse = namedtuple(
    'GraphQLResponse', ('errors', 'ok', 'data'),
)


class Gql:
    """
    Base class for graphQL HTTP queries
    """
    def __init__(
        self, *,
        api: str,
        session: Optional['requests.Session'] = None,
        default_headers: Optional[Mapping[str, Any]] = None,
    ) -> None:
        self.api = api
        self.session = session or requests.Session()
        if default_headers:
            self.session.headers.update(default_headers)

    def generate_result(
        self, result: 'requests.Response'

    ) -> 'GraphQLResponse':
        json_data = result.json()
        return GraphQLResponse(
            errors=json_data.get('errors') or None,
            ok=result.status_code == 200 and not json_data.get('errors', []),
            data=json_data.get('data') or None,
        )

    def send(
        self, *,
        query: str,
        variables: Optional[Mapping[str, Any]] = None,
        headers: Mapping[str, Any] = None,
    ) -> 'GraphQLResponse':

        cleaned_query = normalise_query(query)
        camelcased_variables = to_camelcase(variables) if variables else None
        payload = prepared_payload(
            cleaned_query=cleaned_query,
            camelcased_variables=camelcased_variables,
        )
        all_headers = self.session.headers
        if headers:
            all_headers.update(**headers)
        req = requests.Request(
            url=self.api, method='POST', json=payload, headers=headers
        )
        prepared_request = self.session.prepare_request(req)
        result = self.session.send(prepared_request)
        return self.generate_result(result)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `gql_py` package."""

import json

import pytest
import responses

from gql_py import Gql, normalise_query, to_camelcase


def test_basic_object():
    gql = Gql(api='http://test.com')
    assert gql.api == 'http://test.com'


@responses.activate
def test_headers_are_set():
    responses.add(responses.POST, 'http://test.com/graphql',
                  body='''
                    {"errors": [], "data": {"title": "blog title"}}''',
                  status=200,
                  content_type='application/json')
    gql = Gql(api='http://test.com/graphql')
    gql.send(query='query \{\}', variables={}, headers={'foo': 'bar'})
    assert len(responses.calls) == 1
    assert responses.calls[0].request.url == 'http://test.com/graphql'
    assert 'foo' in responses.calls[0].request.headers.keys()


@responses.activate
def test_headers_are_set_during_object_creation():
    responses.add(responses.POST, 'http://test.com/graphql',
                  body='''
                    {"errors": [], "data": {"title": "blog title"}}''',
                  status=200,
                  content_type='application/json')
    gql = Gql(api='http://test.com/graphql', default_headers={'foo': 'bar'})
    gql.send(query='query {}', variables={})
    assert len(responses.calls) == 1
    assert responses.calls[0].request.url == 'http://test.com/graphql'
    assert 'foo' in responses.calls[0].request.headers.keys()


@responses.activate
def test_good_formatted_data_is_sent():
    responses.add(responses.POST, 'http://test.com/graphql',
                  body='''
                    {"errors": [], "data": {"title": "blog title"}}''',
                  status=200,
                  content_type='application/json')
    gql = Gql(api='http://test.com/graphql')
    query = '''
        query ($someValue: ID){
            concept(id: $someValue){
                name
            }
        }

    '''
    gql.send(
        query=query,
        variables={'some_value': 'bar'},
        headers={'foo': 'bar'}
    )
    http_call = responses.calls[0].request
    json_post_data = json.loads(http_call.body)
    assert json_post_data['variables'] == {'someValue': 'bar'}
    assert json_post_data['query'] == (
        'query ($someValue: ID){ concept(id: $someValue){ name } }'
    )


@responses.activate
def test_good_response():
    responses.add(responses.POST, 'http://test.com/graphql',
                  body='''
                    {"errors": [], "data": {"title": "blog title"}}''',
                  status=200,
                  content_type='application/json')
    gql = Gql(api='http://test.com/graphql')
    result = gql.send(query='query \{\}', variables={}, headers={'foo': 'bar'})
    assert result.errors is None
    assert result.ok
    assert result.data == {'title': 'blog title'}


@responses.activate
def test_bad_response():
    responses.add(responses.POST, 'http://test.com/graphql',
                  body='''
                    {"errors": [{"message": "some error"}], "data": {}}''',
                  status=200,
                  content_type='application/json')
    gql = Gql(api='http://test.com/graphql')
    result = gql.send(query='query \{\}', variables={}, headers={'foo': 'bar'})
    assert result.errors == [{'message': 'some error'}]
    assert not result.ok
    assert result.data is None


@pytest.mark.parametrize('input, output', [
    (
        {'some_variable': 'value'}, {'someVariable': 'value'}
    ),
    (
        {'nested': {'change_me': 'value'}}, {'nested': {'changeMe': 'value'}}
    ),
    (
        {'ignoreMe': 'value'}, {'ignoreMe': 'value'}
    )
])
def test_to_camel_case(input, output):
    assert to_camelcase(input) == output


def test_normalise_query():
    query = '''
        query ($someValue: ID){
            concept(id: $someValue){
                name
            }
        }

    '''
    assert normalise_query(query) == (
        'query ($someValue: ID){ concept(id: $someValue){ name } }'
    )
